/* 10:45 Dec 16, 2012 wzh */

%{ open Type %}
%{ open Ast %}

%token LPAREN RPAREN LBRACKET RBRACKET COLON SEMI COMMA PLUS MINUS TIMES DIVIDE MOD ASSIGN EQ NEQ LT
%token LEQ GT GEQ NOT AND OR ATX ATY ATR ATG ATB ATK IF ELSEIF ELSE SWITCH CASE DEFAULT FOR WHILE
%token CONTINUE BREAK FUNCTION RETURN END VOID INT FLOAT CHAR STRING BOOLEAN POINT COLOR KEY EOF

%token <int> INT_LITERAL
%token <char> CHAR_LITERAL
%token <float> FLOAT_LITERAL
%token <string> STRING_LITERAL
%token <bool> BOOLEAN_LITERAL
%token <int*int> POINT_LITERAL
%token <int*int*int> COLOR_LITERAL
%token <string> KEY_LITERAL
%token <string> ID

%right ASSIGN
%left OR
%left AND
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE MOD
%nonassoc UMINUS NOT ATX ATY ATR ATG ATB ATK

%start program
%type <Ast.program> program

%%

program:
  /* nothing */ { [] }
| program vdecl { Var_decl($2) :: $1 }
| program fdecl { Func_decl($2) :: $1 }	/* in reverse order */

fdecl:
	FUNCTION return_type_specifier ID LPAREN formals_opt RPAREN stmt_list END FUNCTION
		{ {	rtype	= $2;
			fname	= $3;
			formals	= $5;
			body	= List.rev $7 } }

formals_opt:
  /* nothing */			{ [] }
| formal_list			{ List.rev $1 }

formal_list:
  type_specifier ID				{ [($1, $2)] }
| formal_list COMMA type_specifier ID		{ ($3, $4)::$1 }	/* in reverse order */

stmt_list:
  /* nothing */			{ [] }
| stmt_list stmt		{ $2 :: $1 }	/* in reverse order */

stmt:
  expr SEMI									{ Expr($1) }
| RETURN expr SEMI								{ Return($2) }
| RETURN SEMI									{ Return(NoExpr) }
| IF LPAREN expr RPAREN stmt_list END IF					{ If([$3], [Block(List.rev $5)]) }
| IF LPAREN expr RPAREN stmt_list ELSE stmt_list END IF			{ If([$3], [Block(List.rev $5); Block(List.rev $7)]) }
| IF LPAREN expr RPAREN stmt_list elseif_stmt END IF				{ If($3::fst $6, Block(List.rev $5)::snd $6) }
| FOR LPAREN expr_opt SEMI expr_opt SEMI expr_opt RPAREN stmt_list END FOR	{ For($3, $5, $7, Block(List.rev $9)) }
| FOR LPAREN vdecl expr_opt SEMI expr_opt RPAREN stmt_list END FOR		{ ForWithDecl($3, $4, $6, Block(List.rev $8)) }
| WHILE LPAREN expr RPAREN stmt_list END WHILE				{ While($3, Block(List.rev $5)) }
| SWITCH LPAREN expr RPAREN stmt_list case_stmt END SWITCH			{ Switch($3, fst $6, snd $6) }
| SWITCH LPAREN expr RPAREN stmt_list END SWITCH				{ Switch($3, [], []) }	/* stmts before cases will not be executed */
| CONTINUE SEMI								{ Continue }
| BREAK SEMI									{ Break }
| vdecl										{ VarDecl((fst $1, snd $1)) }

init_list:								/* in reverse order */
  init					{ [ $1 ] }
| init_list COMMA init			{ $3 :: $1 }

init:
  ID ASSIGN expr			{ ($1, $3) }
| ID					{ ($1, NoExpr) }

vdecl:
  type_specifier init_list SEMI	{ ($1, List.rev $2) }	

expr:
  ID					{ Id($1) }
| constant				{ Literal($1) }
| ID ASSIGN expr			{ Assign($1, $3) }
| expr PLUS expr			{ Binop($1, Add, $3) }
| expr MINUS expr			{ Binop($1, Sub, $3) }
| expr TIMES expr			{ Binop($1, Mul, $3) }
| expr DIVIDE expr			{ Binop($1, Div, $3) }
| expr MOD expr			{ Binop($1, Mod, $3) }
| expr GT expr				{ Binop($1, Gt, $3) }
| expr LT expr				{ Binop($1, Lt, $3) }
| expr GEQ expr			{ Binop($1, Geq, $3) }
| expr LEQ expr			{ Binop($1, Leq, $3) }
| expr EQ expr				{ Binop($1, Eq, $3) }
| expr NEQ expr			{ Binop($1, Neq, $3) }
| expr AND expr			{ Binop($1, And, $3) }
| expr OR expr				{ Binop($1, Or, $3) }
| MINUS expr %prec UMINUS		{ Unop(Neg, $2) }
| NOT expr				{ Unop(Not, $2) }
| expr ATX				{ Unop(Atx, $1) }
| expr ATY				{ Unop(Aty, $1) }
| expr ATR				{ Unop(Atr, $1) }
| expr ATG				{ Unop(Atg, $1) }
| expr ATB				{ Unop(Atb, $1) }
| expr ATK				{ Unop(Atk, $1) }
| ID LPAREN actuals_opt RPAREN	{ Call($1, $3) }
| LPAREN expr RPAREN			{ $2 }

constant:
  INT_LITERAL				{ Integer($1) }
| CHAR_LITERAL			{ Char($1) }
| STRING_LITERAL			{ String($1) }
| FLOAT_LITERAL			{ Float($1) }
| BOOLEAN_LITERAL			{ Boolean($1) }
| POINT_LITERAL			{ Point(fst $1, snd $1) }
| KEY_LITERAL				{ Key($1) }
| COLOR_LITERAL			{ Color((let fst (a,_,_) = a in fst $1), (let snd (_,b,_) = b in snd $1), (let thd (_,_,c) = c in thd $1)) }

elseif_stmt:
  ELSEIF LPAREN expr RPAREN stmt_list			{ ([$3], [Block(List.rev $5)]) }		
| ELSEIF LPAREN expr RPAREN stmt_list elseif_stmt		{ ($3::fst $6, Block(List.rev $5)::snd $6) }
| ELSEIF LPAREN expr RPAREN stmt_list ELSE stmt_list	{ ([$3], [Block(List.rev $5); Block(List.rev $7)]) }

case_stmt:
  CASE expr COLON stmt_list				{ ([$2], [Block(List.rev $4)]) }
| CASE expr COLON stmt_list case_stmt			{ ($2::fst $5, Block(List.rev $4)::snd $5) }
| DEFAULT COLON stmt_list				{ ([], [Block(List.rev $3)]) }

expr_opt:
  /* nothing */ 			{ NoExpr }
| expr					{ $1 }

actuals_opt:
  /* nothing */ 			{ [] }
| actuals_list				{ List.rev $1 }

actuals_list:
  expr					{ [$1] }
| actuals_list COMMA expr		{ $3 :: $1 }	/* in reverse order */

type_specifier:
  INT					{ Int_T }
| FLOAT				{ Float_T }
| CHAR					{ Char_T }
| STRING				{ String_T }
| BOOLEAN				{ Boolean_T }
| POINT				{ Point_T }
| COLOR				{ Color_T }
| KEY					{ Key_T }

return_type_specifier:
  VOID					{ Void_T }
| type_specifier			{ $1 }
