(* 11:37 Dec 16, 2012 hxy, wzh *)

open Type

type expr_detail =				(* Expressions *)
  Id of string 					(* foo *)
| Literal of constant 				(* 12 $PgDn$ '@' *)
| Assign of string * expr 				(* foo = 823 *)
| Binop of expr * binop * expr				(* x=a+b *)
| Unop of unop * expr					(* !a -t p@x *)
| Call of string * expr list 			(* foo(a,23)) *)
| NoExpr 						(* for(;;) *)
and expr = expr_detail * t * bool			(* Expressions with type and integer that indicates constant expression or not *)

type var_decl = t * string * expr option		(* Variable declaration *)

type stmt = 					(* Statements *)
  Expr of expr 					(* foo = bar + 3; *)
| Block of stmt list * (t * string) list	 	(* statement list and var list *) 
| Return of expr 					(* return 42 *)
| If of expr list * stmt list				(* if(foo==42) ... elseif(foo<42) ... else ... end if *)
| For of expr * expr * expr * stmt 			(* for(i=0;i<10;i=i+1) ... end for *)
| ForWithDecl of var_decl list * expr * expr * stmt	(* for(int i=0;i<10;i=i+1) ... end for *)
| While of expr * stmt 				(* while (i <10) ... end while *)
| Switch of expr * expr list * stmt list		(* switch(x+y) case 1: a=8; b=23; case 2: c=8; d=23; end switch *)
| Continue of int 					(* continue statement with num of locals that need to be popped *)
| Break of int 					(* break statement num of locals that need to be popped *)
| VarDecl of var_decl					(* int a, b = 3, c; *)

type func_decl = {      				(* Function declaration *)
	rtype : t;    					(* return type of the function *)                                                             
	fname : string ;				(* name of the function *)                                                                   
	formals : (t * string) list;			(* types and names of formal arguments *)
	locals : (t * string) list ;			(* local variables *)                     
	body : stmt list ; 				(* body of the function *)                             
}                                                                                                       
  
type program =                                                                                       
	var_decl list * func_decl list 		(* global vars, funcs *)
