(* 03:40 Dec 16, 2012 wzh *)

{ open Parser (* Get the token types *)

let get_point str = 	(* string -> int * int = <fun>, e.g. "{12,345}" -> (12,345) *)
let index = String.index str ',' in
(int_of_string (String.sub str 1 (index-1)), int_of_string (String.sub str (index+1) (String.length str-index-2)))

let get_color str = 	(* string -> int * int * int = <fun>, e.g. "#12,23,34#" -> (12,23,34) *)
let index1 = String.index_from str 0 ',' in
let index2 = String.index_from str (index1+1) ',' in
(int_of_string (String.sub str 1 (index1-1)), int_of_string (String.sub str (index1+1) (index2-index1-1)), int_of_string (String.sub str (index2+1) (String.length str-index2-2)))

let check_color_boundary (r, g, b) =
let list = [r; g; b] in
let checked_list = List.map (fun c -> if c >= 0 && c <= 255 then c else 
raise(Failure("color #" ^ (string_of_int r) ^ ", " ^ (string_of_int g) ^ ", " ^ (string_of_int b) ^ "# out of boundary"))) list
		in (fun [r; g; b] -> (r, g, b)) checked_list

let get_key str = 	(* string -> string = <fun>, e.g. "$Home$" -> "Home" *)
String.sub str 1 (String.length str-2)

let get_ascii str = 	(* string -> string = <fun>, e.g. "'\121'" -> 121 *)
String.sub str 2 3

let get_char n str = 	(* string -> char = <fun>, e.g. "\'\\n\'" -> '\n' *)
let c = String.get str n 
in match c with
  '\\' -> '\\'
| 'b'  -> '\b'
| 't'  -> '\t'
| 'n'  -> '\n'
| 'r'  -> '\r'
| '\'' -> '\''
| '\"' -> '\"'
| _    -> raise(Failure("Illegal escape character : " ^ "\\" ^ (String.make 1 c)))

let rec get_str str tar_str =	(* val get_str : string -> string -> string = <fun>, e.g. "Hello\\n" -> "Hello\n" *)
match str with
  "" -> tar_str
| _  -> if(String.get str 0) = '\\' then
		let c = String.get str 1 in
		if (c >= '0' && c <= '9') then
			get_str (String.sub str 4 ((String.length str)-4)) (tar_str ^ (String.make 1 (let lxm = String.sub str 1 3 in
				try Char.chr (int_of_string lxm) with
					Invalid_argument "Char.chr" -> raise (Failure("Character \\" ^ lxm ^ " out of boundary")))))
		else
			get_str (String.sub str 2 ((String.length str)-2)) (tar_str ^ (String.make 1 (get_char 1 (String.sub str 0 2))))
	else
		get_str (String.sub str 1 ((String.length str)-1)) (tar_str ^ (String.make 1 (String.get str 0)))
		
}

let letter = ['a'-'z' 'A'-'Z']
let digit = ['0'-'9']
let character = ['\b' '\t' '\n' '\r' ' ' '!' '#' '$' '%' '&' '(' ')' '*' '+' ',' '-' '.' '/' ':' ';' '<' '=' '>' '?' '@' '[' ']'
		 '^' '_' '`' '{' '|' '}' '~'] | letter | digit		(*No backslash and quotation marks here *)

let exponent = 'e' ['+' '-']? ['0'-'9']+
let keys = ("a"|"b"|"c"|"d"|"e"|"f"|"g"|"h"|"i"|"j"|"k"|"l"|"m"|"n"|"o"|"p"|"q"|"r"|"s"|"t"|"u"|"v"|"w"|"x"|"y"|"z"
	    |"A"|"B"|"C"|"D"|"E"|"F"|"G"|"H"|"I"|"J"|"K"|"L"|"M"|"N"|"O"|"P"|"Q"|"R"|"S"|"T"|"U"|"V"|"W"|"X"|"Y"|"Z"
	    |"0"|"1"|"2"|"3"|"4"|"5"|"6"|"7"|"8"|"9"|"`"|"~"|"!"|"@"|"#"|"$"|"%"|"^"|"&"|"*"|"("|")"|"-"|"_"|"+"|"="
	    |"["|"{"|"]"|"}"|"\\"|"|"|";"|":"|"\'"|"\""|","|"<"|"."|">"|"/"|"?"|"Enter"|"Space"|"Backspace"|"Esc"|"Tab"|"CapsLk"
	    |"Shift"|"Ctrl"|"Alt"|"Windows"|"Up"|"Down"|"Left"|"Right"|"Delete"|"Home"|"End"|"PgUp"|"PgDn"|"PrtSc"|"ScrLk"
	    |"Insert"|"F1"|"F2"|"F3"|"F4"|"F5"|"F6"|"F7"|"F8"|"F9"|"F10"|"F11"|"F12"|"Null")
let escape = ("\\\\"|"\\b"|"\\t"|"\\n"|"\\r"|"\\\'"|"\\\"")

rule token = parse
[' ' '\t' '\r' '\n'] { token lexbuf }	(* Whitespace *)
| "/*"               { comment lexbuf }	(* Comments *)
| '('        { LPAREN }		(* Punctuation *)
| ')'        { RPAREN }
| '['        { LBRACKET }
| ']'        { RBRACKET }
| ':'        { COLON }
| ';'        { SEMI }
| ','        { COMMA }
| '+'        { PLUS }
| '-'        { MINUS }
| '*'        { TIMES }
| '/'        { DIVIDE }
| "%"        { MOD }
| '='        { ASSIGN }
| "=="       { EQ }
| "!="       { NEQ }
| '<'        { LT }
| "<="       { LEQ }
| ">"        { GT }
| ">="       { GEQ }
| '!'        { NOT }
| "&&"       { AND }
| "||"       { OR }
| "@x"       { ATX }
| "@y"       { ATY }
| "@r"       { ATR }
| "@g"       { ATG }
| "@b"       { ATB }
| "@k"       { ATK }
| "if"       { IF } 			(* Keywords *)
| "elseif"   { ELSEIF }
| "else"     { ELSE }
| "switch"   { SWITCH }
| "case"     { CASE }
| "default"  { DEFAULT }
| "for"      { FOR }
| "while"    { WHILE }
| "continue" { CONTINUE }
| "break"    { BREAK }
| "function" { FUNCTION }
| "return"   { RETURN }
| "end"      { END }
| "void"     { VOID }
| "int"      { INT }
| "float"    { FLOAT }
| "char"     { CHAR }
| "string"   { STRING }
| "boolean"  { BOOLEAN }
| "point"    { POINT }
| "color"    { COLOR }
| "key"      { KEY }
| eof        { EOF }			(* End-of-file *)
| digit+ 									as lxm { INT_LITERAL(int_of_string lxm) }		(* integers *)
| '\'' character '\'' 								as lxm { CHAR_LITERAL(String.get lxm 1) }		(* char *)
| '\'' escape '\''								as lxm { CHAR_LITERAL(get_char 2 lxm) }
| '\'' '\\' digit digit digit '\''						as lxm { CHAR_LITERAL(try Char.chr (int_of_string (get_ascii lxm))
										with Invalid_argument "Char.chr" -> raise 
										(Failure("Character " ^ lxm ^ " out of boundary"))) }
| digit+ ('.' digit* exponent? | exponent) | digit* '.' digit+ exponent? 	as lxm { FLOAT_LITERAL(float_of_string lxm) }	(* float *)
| '\"' (character | escape | '\\' digit digit digit )* '\"'			as lxm { STRING_LITERAL(get_str (get_key lxm) "") }	(* string *)
| "true" | "false"								as lxm { BOOLEAN_LITERAL(bool_of_string lxm) }	(* boolean *)
| '{' digit+ ',' digit+ '}' 							as lxm { POINT_LITERAL(get_point lxm) }		(* point *)
| '#' digit+ ',' digit+ ',' digit+ '#'						as lxm { COLOR_LITERAL(check_color_boundary(get_color lxm)) }		(* color *)
| '$' keys '$' 									as lxm { KEY_LITERAL(get_key lxm) }			(* key *)
| ['a'-'z' 'A'-'Z' '_']['a'-'z' 'A'-'Z' '0'-'9' '_']* 				as lxm { ID(lxm) }
| _ as char { raise (Failure("illegal character " ^ Char.escaped char)) }

and comment = parse
  "*/" { token lexbuf }			(* End-of-comment *)
| _    { comment lexbuf }		(* Eat everything else *)
